#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
from oauthlib.oauth2 import RequestValidator, Server

import iam.domain as iam
from firefly_iaaa.domain.entity.grant import Grant
from firefly_iaaa.domain.service.request_validator import RequestValidator as IRequestValidator


class OauthlibRequestValidator(RequestValidator):
    _registry: ff.Registry = None

    def authenticate_client(self, request, *args, **kwargs):
        u = request.body['username']
        user = self._registry(iam.User).find_one_matching(
            (iam.User.c.email == u) | (iam.User.c.preferred_username == u)
        )
        return self.validate_user(
            request.body['username'], request.body['password'], user.client, request, *args, **kwargs
        )

    def authenticate_client_id(self, client_id, request, *args, **kwargs):
        client = self._get_client(client_id)
        if client:
            request.client = client
            return True
        return False

    def confirm_redirect_uri(self, client_id, code, redirect_uri, client, request, *args, **kwargs):
        client = client or self._get_client(client_id)
        grant = self._registry(Grant).find_one_matching(
            (Grant.c.client_id == client.client_id) & (Grant.c.code == code)
        )
        if not grant:
            return False
        return grant.validate_redirect_uri(redirect_uri)

    def get_default_redirect_uri(self, client_id, request, *args, **kwargs):
        request.client = request.client or self._get_client(client_id)
        return request.client.default_redirect_uri

    def get_default_scopes(self, client_id, request, *args, **kwargs):
        pass

    def get_original_scopes(self, refresh_token, request, *args, **kwargs):
        pass

    def introspect_token(self, token, token_type_hint, request, *args, **kwargs):
        pass

    def invalidate_authorization_code(self, client_id, code, request, *args, **kwargs):
        pass

    def revoke_token(self, token, token_type_hint, request, *args, **kwargs):
        pass

    def save_authorization_code(self, client_id, code, request, *args, **kwargs):
        pass

    def save_bearer_token(self, token, request, *args, **kwargs):
        pass

    def validate_bearer_token(self, token, scopes, request):
        pass

    def validate_client_id(self, client_id, request, *args, **kwargs):
        client = self._get_client(client_id)
        if client:
            return True
        return False

    def validate_code(self, client_id, code, client, request, *args, **kwargs):
        pass

    def validate_grant_type(self, client_id, grant_type, client, request, *args, **kwargs):
        pass

    def validate_redirect_uri(self, client_id, redirect_uri, request, *args, **kwargs):
        request.client = request.client or self._get_client(client_id)
        return request.client.validate_redirect_uri(redirect_uri)

    def validate_refresh_token(self, refresh_token, client, request, *args, **kwargs):
        pass

    def validate_response_type(self, client_id, response_type, client, request, *args, **kwargs):
        return client.validate_response_type(response_type)

    def validate_scopes(self, client_id, scopes, client, request, *args, **kwargs):
        return client.validate_scopes(scopes)

    def validate_user(self, username, password, client, request, *args, **kwargs):
        if client.user.correct_password(password):
            request.user = client.user.email
            return True
        return False

    def get_code_challenge_method(self, code, request):
        pass

    def _get_client(self, client_id: str):
        return self._registry(iam.Client).find(client_id)


class IamRequestValidator(IRequestValidator):
    def __init__(self, validator: OauthlibRequestValidator):
        self._server = Server(validator)

    def validate_pre_auth_request(self, request: ff.Message):
        http_request = request.headers.get('http_request')
        return self._server.validate_authorization_request(
            f'{http_request["headers"]["Host"]}{http_request["url"]}',
            http_request['method'],
            '',
            http_request['headers']
        )

    def validate_post_auth_request(self, request: ff.Message):
        pass

    def create_response(self, request: ff.Message):
        return self._server.create_authorization_response(
            request.headers.get('uri'), request.headers.get('http_method'), request.to_dict(), request.headers
        )



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
from firefly.ui.web.components.form import Form
from firefly.ui.web.components.layouts.default import Crud, menu_item, crud
from firefly.ui.web.js_libs.mithril import m
from firefly.ui.web.plugins import add_menu_item, add_route
from firefly.ui.web.polyfills import *  # __:skip

from firefly_iaaa.domain.entity.user import User
from firefly_iaaa.domain.entity.client import Client

add_menu_item(m('div.ff-title', 'IAM'))
add_menu_item(m(menu_item('Users', icon='solid/users', route='/iam/users')))
add_menu_item(m(menu_item('Clients', icon='solid/mobile-alt', route='/iam/clients')))

crud('iam.User', User, '/iam/users')
crud('iam.Client', Client, '/iam/clients', {
    'fields': [
        'name',
        'grant_type',
        'response_type',
        'scopes',
        'default_redirect_uri',
        'redirect_uris',
        'allowed_response_types',
    ]
})

# add_route('/iam/users', Crud('iam.User', User, '/iam/users'))
# add_route('/iam/clients', Crud('iam.Client', Client, '/iam/clients', {
#     'fields': [
#         'name',
#         'grant_type',
#         'response_type',
#         'scopes',
#         'default_redirect_uri',
#         'redirect_uris',
#         'allowed_response_types',
#     ]
# }))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List

import firefly as ff
from .tenant import Tenant

authorization_code = 'Authorization Code'
implicit = 'Implicit'
resource_owner_password_credentials = 'Resource Owner Password Credentials'
client_credentials = 'Client Credentials'


def response_type_choices(client_dto: dict):
    if client_dto['grant_type'] == authorization_code:
        return 'code token', 'code id_token', 'code token id_token'
    if client_dto['grant_type'] == implicit:
        return 'id_token token', 'id_token'

    return ()


class Client(ff.AggregateRoot):
    id: str = ff.id_()
    external_id: str = ff.optional(index=True)
    name: str = ff.required()
    grant_type: str = ff.required(validators=[ff.IsOneOf((
        authorization_code, implicit, resource_owner_password_credentials, client_credentials
    ))])
    response_type: str = ff.optional(validators=[ff.IsOneOf(response_type_choices)])
    scopes: str = ff.required()
    default_redirect_uri: str = ff.optional()
    redirect_uris: List[str] = ff.list_()
    allowed_response_types: List[str] = ff.list_(validators=[ff.IsOneOf(('code', 'token'))])
    tenant: Tenant = ff.optional(index=True)

    def validate_redirect_uri(self, redirect_uri: str):
        return redirect_uri in self.redirect_uris

    def validate_response_type(self, response_type: str):
        return response_type in self.allowed_response_types

    def validate_scopes(self, scopes: List[str]):
        for scope in scopes:
            if scope not in self.scopes:
                return False
        return True



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from firefly.ui.web.components.firefly_logo import FireflyLogo, FireflyIcon
from firefly.ui.web.polyfills import *  # __:skip
from firefly.ui.web.js_libs.mithril import m


class MainMenu:
    PAGES = ('clients', 'users')

    def __init__(self, exclude: List[str] = None):
        self._exclude = exclude or []

    def _menu_item(self, page: str):
        if page in self._exclude:
            return None

        return m(
            'div.pl-3.py-1',
            m(m.route.Link, {'href': f'/{page}'}, page)
        )

    def view(self):
        return m(
            'div.flex.flex-col.justify-start', [
                m(FireflyLogo()),
                list(map(self._menu_item, self.PAGES))
            ]
        )



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from firefly.ui.web.polyfills import *  # __:skip
from firefly.ui.web.js_libs.mithril import m
from firefly.ui.web import bus, Form
from firefly_iaaa_web.state import state

from firefly_iaaa.domain.entity.user import User


# __pragma__('opov')
class UserForm:
    def __init__(self):
        self.form = Form(
            User(),
            {
                'onsubmit': lambda data: bus.invoke('iam.CreateUser', User(**data).to_dict()),
                'fieldsets': [
                    {
                        'fields': ['given_name', 'middle_name', 'family_name', 'name', 'nickname', 'gender', 'birthdate'],
                        'legend': 'Personal',
                    },
                    {
                        'fields': ['email', 'email_verified', 'phone_number', 'phone_number_verified'],
                        'legend': 'Contact',
                    },
                    {
                        'fields': ['address'],
                        'legend': 'Address',
                    },
                    {
                        'fields': ['preferred_username', 'locale', 'picture', 'profile', 'website', 'zoneinfo'],
                        'legend': 'Account',
                    }
                ]
            }
        )

    def view(self):
        return [
            m('h1', 'User Form'),
            m(self.form),
            m('h1', 'Users'),
            m('ul', map(lambda u: m('li', u.given_name), state.users()))
        ]
# __pragma__('noopov')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.


from firefly.ui.web.polyfills import *  # __:skip
from firefly.ui.web.js_libs.mithril import m, Stream

from typing import Any, Callable

from firefly import Query, MessageFactory, Message
from firefly.ui.web.bus import bus


def __pragma__(*args):
    pass

# __pragma__('js', '{}', "var inflection = require('inflection');")

# __pragma__('opov')
# __pragma__('kwargs')

import iam

mf = MessageFactory()


class QueryStream:
    _bus = bus

    def __init__(self, query: Query = None, default: Any = None, fetch: str = 'lazy', handlers: dict = None,
                 refresh_on: list = None):
        self._initialized = False
        self._query = query
        self._stream = Stream(default or [])
        self._stream.map(lambda: m.redraw())

        def _set_stale(val):
            self._stale = False
        self._stream.map(_set_stale)

        self._aggregate = inflection.singularize(query.__class__.__name__)
        self._stale = False

        if handlers is not None:
            for command, handler in handlers.items():
                def mw(message: Message, next_: Callable):
                    context, cmd = str(command).split('.')
                    if message.get_context() == context and message.__class__.__name__ == cmd:
                        self._stream(handler(message, self._stream()))
                    return next_(message)
                self._bus.insert_command_handler(1, mw)

        if refresh_on is not None:
            for command in refresh_on:
                def mw(message: Message, next_: Callable):
                    context, cmd = str(command).split('.')
                    ret = next_(message)
                    if message.get_context() == context and message.__class__.__name__ == cmd:
                        if hasattr(ret, 'then'):
                            return ret.then(lambda _: self.refresh())
                    return ret
                self._bus.insert_command_handler(1, mw)

        if fetch == 'eager':
            self._execute_query()

    def __call__(self, value: Callable = None):
        if value is None:
            if not self._initialized:
                self._execute_query()
            return self._stream()
        return self._stream(value)

    def __getattr__(self, item):
        return getattr(self._stream, item)

    def matches_aggregate(self, aggregate: str):
        return self._aggregate == aggregate

    def refresh(self):
        self._execute_query()

    def is_stale(self):
        return self._stale

    def _execute_query(self):
        self._stale = True
        self._bus.request(self._query).then(self._stream)
        self._initialized = True


class State:
    def __init__(self):
        self.users = QueryStream(
            query=mf.query('iam.Users'),
            default=[],
            handlers={
                'iam.CreateUser': lambda c, s: s + [iam.User(**c.to_dict())]
            },
        )


state = State()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
from datetime import datetime, date
from typing import List

import firefly as ff

from firefly_iaaa.domain.value_object.address import Address
from .role import Role

# __pragma__('skip')
import bcrypt
import uuid
# __pragma__('noskip')


class User(ff.AggregateRoot):
    # OpenID standard fields
    sub: str = ff.id_(validators=[ff.HasLength(36)])
    name: str = ff.optional(str)
    given_name: str = ff.optional(str)
    family_name: str = ff.optional(str)
    middle_name: str = ff.optional(str)
    nickname: str = ff.optional(str)
    preferred_username: str = ff.optional(str)
    profile: str = ff.optional(str)
    picture: str = ff.optional(str)
    website: str = ff.optional(str)
    email: str = ff.optional(str, validators=[ff.IsValidEmail()])
    email_verified: bool = ff.optional(bool, default=False)
    gender: str = ff.optional(str, validators=[ff.IsOneOf(('Male', 'Female'))])
    birthdate: date = ff.optional(date)
    zoneinfo: str = ff.optional(str)
    locale: str = ff.optional(str)
    phone_number: str = ff.optional(str)
    phone_number_verified: bool = ff.optional(bool, default=False)
    address: Address = ff.optional(Address)
    updated_at: datetime = ff.now()

    # Custom fields
    created_at: datetime = ff.now()
    deleted_at: datetime = ff.optional(datetime)
    password_hash: str = ff.optional(str, length=32, internal=True)
    salt: str = ff.hidden(internal=True)
    roles: List[Role] = ff.list_()
    tenant: str = ff.optional()

    # __pragma__('skip')
    @classmethod
    def create(cls, **kwargs):
        if 'email' in kwargs:
            kwargs['email'] = str(kwargs['email']).lower()

        try:
            kwargs['salt'] = bcrypt.gensalt()
            kwargs['password_hash'] = User._hash_password(kwargs['password'], kwargs['salt'])
        except KeyError:
            raise ff.MissingArgument('password is a required field for User::create()')
        return cls(**ff.build_argument_list(kwargs, cls))

    @classmethod
    def _hash_password(cls, password: str, salt: str):
        return bcrypt.hashpw(password.encode('utf-8'), salt).decode('utf-8')

    def correct_password(self, password: str):
        return self.password_hash == User._hash_password(password, self.salt)
    # __pragma__('noskip')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from firefly.ui.web.components.layouts.default import compose, default_layout
from firefly.ui.web.js_libs.mithril import m
from firefly.ui.web.polyfills import *  # __:skip

# from firefly_iaaa_web.components.clients_page import ClientsPage
# from firefly_iaaa_web.components.main_menu import MainMenu

m.route.prefix = ''


def counter():
    count = 0

    def increment():
        nonlocal count
        count += 1
        console.log(count)

    def decrement():
        nonlocal count
        count -= 1

    return {
        'view': lambda vnode: m('div', [
            m('p', f'count: {count}'),
            m('button', {'onclick': increment}, 'Increment'),
            m('button', {'onclick': decrement}, 'Decrement'),
        ])
    }


def main_content():
    return {
        'view': lambda: m('div.flex.flex-col', [
            m(counter),
            m(counter),
        ])
    }


def custom_header():
    return {
        'view': lambda: m('div', 'bleck')
    }


m.route(document.body, '/', {
    '/': compose(default_layout, main_content),
})

"""
__pragma__('js', '{}', '''
if (module.hot) {
  module.hot.accept();
}
''')
"""

